@@ -9,7 +9,7 @@
 #endif
 
 #define Proxy__WRAPPED_REPLACE_OR_RETURN_NULL(object) \
-    if (PyObject_IsInstance(object, (PyObject *)&Proxy_Type)) { \
+    if (PyObject_TypeCheck(object, &Proxy_Type)) { \
         object = Proxy__ensure_wrapped((ProxyObject *)object); \
         if (!object) return NULL; \
     }
@@ -37,6 +37,7 @@ PyTypeObject Proxy_Type;
 /* ------------------------------------------------------------------------- */
 
 static PyObject *identity_ref = NULL;
+static PyObject *await_ref = NULL;
 static PyObject *
 identity(PyObject *self, PyObject *value)
 {
@@ -222,6 +223,35 @@ static PyObject *Proxy_str(ProxyObject *self)
 
 /* ------------------------------------------------------------------------- */
 
+static PyObject *Proxy_fspath(ProxyObject *self)
+{
+    PyObject *method = NULL;
+    PyObject *fspath = NULL;
+
+    Proxy__ENSURE_WRAPPED_OR_RETURN_NULL(self);
+
+    if (PyUnicode_Check(self->wrapped) || PyBytes_Check(self->wrapped)) {
+        Py_INCREF(self->wrapped);
+        return self->wrapped;
+    }
+
+    method = PyObject_GetAttrString(self->wrapped, "__fspath__");
+
+    if (!method) {
+        PyErr_Clear();
+        Py_INCREF(self->wrapped);
+        return self->wrapped;
+    }
+
+    fspath = PyObject_CallFunctionObjArgs(method, NULL);
+
+    Py_DECREF(method);
+
+    return fspath;
+}
+
+/* ------------------------------------------------------------------------- */
+
 static PyObject *Proxy_add(PyObject *o1, PyObject *o2)
 {
     Proxy__WRAPPED_REPLACE_OR_RETURN_NULL(o1);
@@ -428,8 +458,7 @@ static PyObject *Proxy_oct(ProxyObject *self)
 
     if ((nb = self->wrapped->ob_type->tp_as_number) == NULL ||
         nb->nb_oct == NULL) {
-        PyErr_SetString(PyExc_TypeError,
-                   "oct() argument can't be converted to oct");
+        PyErr_SetString(PyExc_TypeError, "oct() argument can't be converted to oct");
         return NULL;
     }
 
@@ -448,8 +477,7 @@ static PyObject *Proxy_hex(ProxyObject *self)
 
     if ((nb = self->wrapped->ob_type->tp_as_number) == NULL ||
         nb->nb_hex == NULL) {
-        PyErr_SetString(PyExc_TypeError,
-                   "hex() argument can't be converted to hex");
+        PyErr_SetString(PyExc_TypeError, "hex() argument can't be converted to hex");
         return NULL;
     }
 
@@ -825,8 +853,7 @@ static PyObject *Proxy_dir(
 
 /* ------------------------------------------------------------------------- */
 
-static PyObject *Proxy_enter(
-        ProxyObject *self, PyObject *args, PyObject *kwds)
+static PyObject *Proxy_enter(ProxyObject *self)
 {
     PyObject *method = NULL;
     PyObject *result = NULL;
@@ -838,7 +865,7 @@ static PyObject *Proxy_enter(
     if (!method)
         return NULL;
 
-    result = PyObject_Call(method, args, kwds);
+    result = PyObject_CallObject(method, NULL);
 
     Py_DECREF(method);
 
@@ -1053,6 +1080,18 @@ static int Proxy_set_annotations(ProxyObject *self,
 
 /* ------------------------------------------------------------------------- */
 
+static PyObject *Proxy_get_resolved(
+        ProxyObject *self)
+{
+    PyObject *result;
+
+    result = self->wrapped ? Py_True : Py_False;
+    Py_INCREF(result);
+    return result;
+}
+
+/* ------------------------------------------------------------------------- */
+
 static PyObject *Proxy_get_wrapped(
         ProxyObject *self)
 {
@@ -1198,6 +1237,107 @@ static PyObject *Proxy_call(
 
 /* ------------------------------------------------------------------------- */;
 
+#if PY_MAJOR_VERSION >= 3
+
+static PyObject *Proxy_aenter(ProxyObject *self)
+{
+    PyObject *method = NULL;
+    PyObject *result = NULL;
+
+    Proxy__ENSURE_WRAPPED_OR_RETURN_NULL(self);
+
+    method = PyObject_GetAttrString(self->wrapped, "__aenter__");
+
+    if (!method)
+        return NULL;
+
+    result = PyObject_CallObject(method, NULL);
+
+    Py_DECREF(method);
+
+    return result;
+}
+
+/* ------------------------------------------------------------------------- */
+
+static PyObject *Proxy_aexit(
+        ProxyObject *self, PyObject *args, PyObject *kwds)
+{
+    PyObject *method = NULL;
+    PyObject *result = NULL;
+
+    Proxy__ENSURE_WRAPPED_OR_RETURN_NULL(self);
+
+    method = PyObject_GetAttrString(self->wrapped, "__aexit__");
+
+    if (!method)
+        return NULL;
+
+    result = PyObject_Call(method, args, kwds);
+
+    Py_DECREF(method);
+
+    return result;
+}
+
+/* ------------------------------------------------------------------------- */
+
+static PyObject *Proxy_await(ProxyObject *self)
+{
+    Proxy__ENSURE_WRAPPED_OR_RETURN_NULL(self);
+
+    return PyObject_CallFunctionObjArgs(await_ref, self->wrapped, NULL);
+}
+
+/* ------------------------------------------------------------------------- */;
+
+static PyObject *Proxy_aiter(ProxyObject *self)
+{
+    Proxy__ENSURE_WRAPPED_OR_RETURN_NULL(self);
+
+    unaryfunc meth = NULL;
+    PyObject *wrapped = self->wrapped;
+    PyTypeObject *type = Py_TYPE(wrapped);
+
+    if (type->tp_as_async != NULL) {
+        meth = type->tp_as_async->am_aiter;
+    }
+
+    if (meth != NULL) {
+        return (*meth)(wrapped);
+    }
+
+    PyErr_Format(PyExc_AttributeError, "'%.100s' object has no attribute '__aiter__'", type->tp_name);
+    return NULL;
+}
+
+/* ------------------------------------------------------------------------- */;
+
+static PyObject *Proxy_anext(ProxyObject *self)
+{
+    Proxy__ENSURE_WRAPPED_OR_RETURN_NULL(self);
+
+
+    unaryfunc meth = NULL;
+    PyObject *wrapped = self->wrapped;
+    PyTypeObject *type = Py_TYPE(wrapped);
+
+    if (type->tp_as_async != NULL) {
+        meth = type->tp_as_async->am_anext;
+    }
+
+    if (meth != NULL) {
+        return (*meth)(wrapped);
+    }
+
+    PyErr_Format(PyExc_TypeError, "'%.100s' is missing the __anext__ method", type->tp_name);
+    return NULL;
+}
+
+#endif
+
+/* ------------------------------------------------------------------------- */;
+
 static PyNumberMethods Proxy_as_number = {
     (binaryfunc)Proxy_add,                  /*nb_add*/
     (binaryfunc)Proxy_subtract,             /*nb_subtract*/
@@ -1270,10 +1410,17 @@ static PyMappingMethods Proxy_as_mapping = {
     (objobjargproc)Proxy_setitem, /*mp_ass_subscript*/
 };
 
+#if PY_MAJOR_VERSION >= 3
+static PyAsyncMethods Proxy_as_async = {
+    (unaryfunc)Proxy_await, /* am_await */
+    (unaryfunc)Proxy_aiter, /* am_aiter */
+    (unaryfunc)Proxy_anext, /* am_anext */
+};
+#endif
+
 static PyMethodDef Proxy_methods[] = {
     { "__dir__",    (PyCFunction)Proxy_dir, METH_NOARGS, 0 },
-    { "__enter__",  (PyCFunction)Proxy_enter,
-                    METH_VARARGS | METH_KEYWORDS, 0 },
+    { "__enter__",  (PyCFunction)Proxy_enter, METH_NOARGS, 0 },
     { "__exit__",   (PyCFunction)Proxy_exit,
                     METH_VARARGS | METH_KEYWORDS, 0 },
     { "__getattr__", (PyCFunction)Proxy_getattr,
@@ -1282,8 +1429,12 @@ static PyMethodDef Proxy_methods[] = {
     { "__reversed__", (PyCFunction)Proxy_reversed, METH_NOARGS, 0 },
     { "__reduce__", (PyCFunction)Proxy_reduce, METH_NOARGS, 0 },
     { "__reduce_ex__", (PyCFunction)Proxy_reduce, METH_O, 0 },
+    { "__fspath__", (PyCFunction)Proxy_fspath, METH_NOARGS, 0 },
 #if PY_MAJOR_VERSION >= 3
     { "__round__",  (PyCFunction)Proxy_round, METH_NOARGS, 0 },
+    { "__aenter__", (PyCFunction)Proxy_aenter, METH_NOARGS, 0 },
+    { "__aexit__",  (PyCFunction)Proxy_aexit,
+                    METH_VARARGS | METH_KEYWORDS, 0 },
 #endif
     { NULL, NULL },
 };
@@ -1305,6 +1456,8 @@ static PyGetSetDef Proxy_getset[] = {
                             (setter)Proxy_set_wrapped, 0 },
     { "__factory__",        (getter)Proxy_get_factory,
                             (setter)Proxy_set_factory, 0 },
+    { "__resolved__",       (getter)Proxy_get_resolved,
+                            NULL, 0 },
     { NULL },
 };
 
@@ -1318,7 +1471,11 @@ PyTypeObject Proxy_Type = {
     0,                              /*tp_print*/
     0,                              /*tp_getattr*/
     0,                              /*tp_setattr*/
+#if PY_MAJOR_VERSION >= 3
+    &Proxy_as_async,            /* tp_as_async */
+#else
     0,                              /*tp_compare*/
+#endif
     (unaryfunc)Proxy_repr,          /*tp_repr*/
     &Proxy_as_number,               /*tp_as_number*/
     &Proxy_as_sequence,             /*tp_as_sequence*/
@@ -1400,6 +1557,17 @@ moduleinit(void)
         return NULL;
     Py_INCREF(identity_ref);
 
+#if PY_MAJOR_VERSION >= 3
+    PyObject *utils_module = PyImport_ImportModule("lazy_object_proxy.utils");
+    if (utils_module == NULL)
+        return NULL;
+
+    await_ref = PyObject_GetAttrString(utils_module, "await_");
+    Py_DECREF(utils_module);
+    if (await_ref == NULL)
+        return NULL;
+#endif
+
     Py_INCREF(&Proxy_Type);
     PyModule_AddObject(module, "Proxy",
             (PyObject *)&Proxy_Type);
